@import url("./sphinxdoc.css");

body {
    background-color: #E5E2E2;
}

div.sphinxsidebar h4, div.sphinxsidebar h3 {
    background-color: #2F83C8;
}

h1 {
    color: #2F83C8;
}

div.footer {
    background-color: #CEC890;
}


div.body {
    background-color: white;
}
